/*
 *  typespeed - measures your typing speed
 *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
 *                          & Jaakko Manelius  <jman@iki.fi>
 *  Copyright (C) 2006   Tobias Stoeckmann  <tobias@bugol.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * network.c - network for typespeed
 *
 * thanks goto Steinar H. Gunderson <sgunderson@bigfoot.com>
 * for the basic functions which we got from him
 */

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <curses.h>
#include <err.h>
#include <errno.h>
#include <limits.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "typespeed.h"
#include "version.h"

extern int	ttss;

extern int	 addnewword(char *);
extern void	 endcursestuff(void);
extern int	 r(int);
extern void	*xmalloc(size_t);

void		 closenetwork(void);
int		 initnetwork(char *, int);
static int	 netident(int, int, int);
int		 netrecvword(int);
void		 netsendword(int typesock, char *);
void		 netswapscore(int, struct stats *, struct stats *);
void		 netwaitgame(int);

void
closenetwork(void)
{
	close(ttss);
	ttss = 0;
}

int
initnetwork(char *serv, int graph)
{
	struct sockaddr_in addr;
	int i, one, sock, ss;

	one = 1;

	if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
		if (graph)
			endcursestuff();
		err(1, "socket()");
	}

	bzero((char *)&addr, sizeof(addr));
	addr.sin_family = PF_INET;
	addr.sin_port = htons(opt.port);

	if (opt.client) {
		if (!inet_aton(serv, &addr.sin_addr)) {
			struct hostent *he;

			if (graph) {
				mvprintw(10, 30, "Looking up %s...", serv);
				refresh();
			} else
				printf("Looking up %s...\n", serv);

			if ((he = gethostbyname(serv)) == NULL) {
				if (graph)
					endcursestuff();
				err(1, "gethostbyname(%s)", serv);
			}
			bcopy(he->h_addr, (char *)&addr.sin_addr, he->h_length);
		}

		if (graph) {
			mvaddstr(10,30, "Connecting to typespeed server at");
			mvprintw(11,30, "%s (port %d)...", inet_ntoa(addr.sin_addr), opt.port);
			refresh();
		} else
			printf("Connecting to typespeed server at %s (port %d)...\n", inet_ntoa(addr.sin_addr),opt.port);

		if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
			if (graph)
				endcursestuff();
			err(1, "connect(%s)", serv);
		}

		if ((i = netident(sock, 0, graph)))
			return i;
	} else {
		addr.sin_addr.s_addr = INADDR_ANY;

		if ((ss = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
			if (graph)
				endcursestuff();
			err(1, "socket()");
		}

		setsockopt(ss, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
		if (bind(ss, (struct sockaddr *)&addr, sizeof(struct sockaddr)) == -1) {
			if (graph)
				endcursestuff();
			err(1, "bind()");
		}

		if (listen(ss, 1) == -1) {
			if (graph)
				endcursestuff();
			err(1, "listen()");
		}

		if (graph) {
			mvaddstr(10, 30, "Waiting for connections...");
			refresh();
		} else
			puts("Waiting for connections...");

		if ((sock = accept(ss, INADDR_ANY, NULL)) == -1) {
			if (graph)
				endcursestuff();
			err(1, "accept()");
		}

		ttss = ss;

		if ((i = netident(sock, ss, graph))) {
			close(ss);
			return i;
		}
	}

	ioctl(sock, FIONBIO, &one);

	return sock;
}

static int
netident(int typesock, int tss, int graph)
{
	char buf[BUFSIZ];
	int i, retval;

	i = snprintf(buf, sizeof(buf), "TYPESPEED %s\n", TVERSION);
	if (i < 10 || (size_t)i > sizeof(buf) - 1) {
		endcursestuff();
		errx(1, "netident: malformed version");
	}

	if ((i = send(typesock, buf, strlen(buf), 0)) == -1) {
		endcursestuff();
		err(1, "send()");
	}
	if ((size_t)i != strlen(buf)) {
		endcursestuff();
		errx(1, "%d out of %zd bytes sent", i, strlen(buf));
	}

	memset(buf, 0, sizeof(buf));

	if (graph) {
		mvaddstr(10, 30, "Waiting for other party to identify...");
		refresh();
	} else {
		puts("Waiting for other party to identify...");
	}

	i = recv(typesock, buf, sizeof(buf), MSG_PEEK);
	buf[sizeof(buf) - 1] = '\0';

	if (i == -1 && errno != EWOULDBLOCK) {
		if (graph)
			endcursestuff();
		err(1, "recv()");
	}

	retval = -69;

	if (i > 0) {
		recv(typesock, buf, strlen(buf) + 1, 0);
		if (!strncmp(buf, "TYPESPEED ", strlen("TYPESPEED "))) {
			if (graph) {
				mvprintw(10, 30, "Remote is using: %s", buf);
				refresh();
			} else
				printf("Remote is using: %s\n",buf);
			retval = 0;
		} else {
			if (graph) {
				mvaddstr(10, 30, "Remote is not TYPESPEED");
				refresh();
			} else
				puts("Remote is not TYPESPEED");
			close(tss);
		}
	}

	return retval;
}

void
netswapscore(int typesock, struct stats *stat, struct stats *stat2)
{
	char buf[BUFSIZ];
	char *number, *p;
	int i;

	i = snprintf(buf, sizeof(buf), "SCORE: %d %f %f %f\n", stat->score, stat->speed, stat->totalspeed, stat->ratio);
	if (i < 11 || (size_t)i > sizeof(buf) - 1) {
		endcursestuff();
		errx(1, "netswapscore: malformed line");
	}

	if ((i = send(typesock, buf, strlen(buf), 0)) == -1) {
		endcursestuff();
		err(1, "send()");
	}
	if ((size_t)i != strlen(buf)) {
		endcursestuff();
		errx(1, "%d out of %zd bytes sent", i, strlen(buf));
	}
	memset(buf, 0, sizeof(buf));

	for (;;) {
		i = recv(typesock, buf, sizeof(buf), MSG_PEEK);
		buf[sizeof(buf) - 1] = '\0';

		if (i == -1 && errno != EWOULDBLOCK) {
			endcursestuff();
			err(1, "recv()");
		}
		if (i > 0) {
			recv(typesock, buf, strlen(buf) + 1, 0);
		        if (!strncmp(buf, "SCORE: ", strlen("SCORE: ")))
				break;
		}
	}

	memset(stat2, 0, sizeof(stat2));

	number = p = buf + strlen("SCORE: ");
	if ((p = strchr(p, ' ')) == NULL)
		return;
	*p = '\0';
	stat2->score = strtol(number, NULL, 10);

	number = ++p;
	if ((p = strchr(p, ' ')) == NULL)
		return;
	*p = '\0';
	stat2->speed = (float)atof(number);

	number = ++p;
	if ((p = strchr(p, ' ')) == NULL)
		return;
	*p = '\0';
	stat2->totalspeed = (float)atof(number);

	number = ++p;
	stat2->ratio = (float)atof(number);
}

void
netwaitgame(int typesock)
{
	char buf[BUFSIZ];
	int i;

	i = snprintf(buf, sizeof(buf), "TYPESPEED %s\n", TVERSION);
	if (i < 10 || (size_t)i > sizeof(buf) - 1) {
		endcursestuff();
		errx(1, "netwaitgame: malformed line");
	}

	if ((i = send(typesock, buf, strlen(buf), 0)) == -1) {
		endcursestuff();
		err(1, "send()");
	}
	if ((size_t)i != strlen(buf)) {
		endcursestuff();
		errx(1, "%d out of %zd bytes sent", i, strlen(buf));
	}

	for (;;) {
		i = recv(typesock, buf, sizeof(buf), MSG_PEEK);
		buf[sizeof(buf) - 1] = '\0';

		if (i == -1 && errno != EWOULDBLOCK) {
			endcursestuff();
			err(1, "recv()");
		}
		if (i > 0) {
			recv(typesock, buf, strlen(buf) + 1, 0);
                	if (!strncmp(buf, "TYPESPEED", strlen("TYPESPEED")))
				break;
		}
	}
}

int
netrecvword(int typesock)
{
	char buf[BUFSIZ], *j;
	int retval;
	ssize_t i;

	retval = 0;

	i = recv(typesock, buf, sizeof(buf), MSG_PEEK);
	buf[sizeof(buf) - 1] = '\0';

	if (i == -1 && errno != EWOULDBLOCK) {
		endcursestuff();
		err(1, "recv()");
	}
	if (i > 0) {
		recv(typesock, buf, strlen(buf) + 1, 0);

		if ((j = strchr(buf, '\n')) != NULL) {
			*j = '\0';
			if (!strncmp(buf, " ENDOFTHEGAME", strlen(" ENDOFTHEGAME")))
				retval = 2;
			else
				addnewword(buf);
		}
	}

	return retval;
}

void
netsendword(int typesock, char *wordstring)
{
	char buf[BUFSIZ];
	int i;

	i = snprintf(buf, sizeof(buf), "%s\n", wordstring);
	if (i < 1 || (size_t)i > sizeof(buf) - 1) {
		endcursestuff();
		errx(1, "netsendword: malformed word %s", wordstring);
	}

	if ((i = send(typesock, buf, strlen(buf), 0)) == -1) {
		endcursestuff();
		err(1, "send()");
	}
	if ((size_t)i != strlen(buf)) {
		endcursestuff();
		errx(1, "%d out of %zd bytes sent", i, strlen(buf));
	}
}

