/*
 *  typespeed - measures your typing speed
 *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
 *                          & Jaakko Manelius  <jman@iki.fi>
 *  Copyright (C) 2006   Tobias Stoeckmann  <tobias@bugol.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * typespeed.c - the main code
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <ctype.h>
#include <curses.h>
#include <err.h>
#include <errno.h>
#include <getopt.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "typespeed.h"
#include "version.h"

#define DEFAULT_CHEAT		0
#define DEFAULT_CLIENT		0
#define DEFAULT_COLORS		1
#define DEFAULT_PORT		6025
#define DEFAULT_SERVER		0
#define DEFAULT_TRAINING	0
#define DEFAULT_TRAININGSPEED	2

extern int	 addtop10(char *, struct stats *);
extern int	 choosewordfile(int);
extern void	 closenetwork(void);
extern void	 drawmenu(void);
extern void	 drawstatus(int);
extern void	 drawscreen(void);
extern void	 endcursestuff(void);
extern void	 freewords(void);
extern int	 initnetwork(char *, int);
extern void	 initcursestuff(void);
extern int	 level(int);
extern void	 loadhighscores(char *);
extern void	 makescorefiles(void);
extern void	 multipmenu(void);
extern int	 netrecvword(int);
extern void	 netsendword(int, char *);
extern void	 netswapscore(int, struct stats *, struct stats *);
extern void	 netwaitgame(int);
extern void	 optionmenu(void);
extern void	 pressanykey(int, int);
extern int	 r(int);
extern void	 readconfig(void);
extern void	 showhighscores(void);
extern void	 tellstory(void);
extern clock_t	 timenow(void);
extern int	 typorankkaus(float);
extern void	*xmalloc(size_t);
extern void	 xcolor_set(short);

int			 addnewword(char *);
static void		 clearword(int, int, size_t);
static void		 cwords(void);
static void		 movewords(void);
static inline void	 parseinput(int *, char *, clock_t *, int *, int *, int *);
static int		 play(void);
static void		 usage(char *);

/* globals */
char *rankki[11] =
{"None", "Loser", "ADP-Pro", "NoGood", "Average",
 "Good", "VeryGood", "Pro", "3l33t", "*(GOD)*", "Computer"};

char *typorank[12] =
{"XMunkki", "Alien", "Secretary", "Human", "IT Person", "Handicap",
 "Monkey", "Pencil", "T-Bone", "E-Typo", "TypOmatiC", "TypoKing"};

struct stats	 best;
int		 misses;
struct stats	 other;
float		 rate;
int		 ttss;
int		 typesocket = 0;
char		*usedwordfile;
char		*word[WORD_MAX];
int		 wordcount = 0;
int		 wordpos[22];
char		 wordstring[22][20];

struct option options[] = {
	{"nocolors", no_argument, &opt.usecolors, 0},
	{"cheat", no_argument, &opt.cheat, 1},
	{"training", optional_argument, NULL, 't'},
	{"client", required_argument, NULL, 'o'},
	{"server", no_argument, &opt.server, 1},
	{"port", required_argument, NULL, 'p'},
	{"makescores", no_argument, NULL, 'm'},
	{"help", no_argument, NULL, 'h'},
	{0, 0, 0, 0}
};

/*
 * Insert a new word into wordstring and wordpos, i.e. one more word
 * on the screen.
 *
 * If newword points to a string, newword will be inserted.
 * If newword is NULL, a random word out of word will be inserted.
 * In both cases, duplicates are not allowed. Should newword point
 * to duplicated string or string that cannot be typed with current
 * locale, no string will be inserted.
 *
 * Returns TRUE on success or FALSE on failure.
 *
 * XXX: Can result in a veeery long loop when there are many words on
 *      field and only a few words in wordlist (no duplicates allowed).
 */
int
addnewword(char *newword)
{
	char *myword, *p;
	int count, dup, slot;
	int freeslot[22];
	size_t i;

	if (newword != NULL) {
		if (strlen(newword) > 19)
			return FALSE;
		for (p = newword; *p != '\0'; p++)
			if (!isprint(*p) || isspace(*p))
				return FALSE;
	}

	/* take a random free slot */
	for (count = 0, i = 0; i < 22; i++)
		if (!wordpos[i])
			freeslot[count++] = i;

	if (!count)
		/* the player has enough trouble... */
		return FALSE;

	slot = freeslot[r(count)];

	if ((myword = newword) == NULL)
		myword = word[r(wordcount)];

	do {
		for (dup = 0, i = 0; i < 22; i++)
			if (!strcmp(wordstring[i], myword)) {
				if (newword != NULL)
					return FALSE;
				dup = 1;
				myword = word[r(wordcount)];
				break;
			}
	} while (dup);

	(void)strncpy(wordstring[slot], myword, sizeof(wordstring[slot]) - 1);
	wordpos[slot] = 1;
	return TRUE;
}

/* Writes "length" blanks at position x y. */
static void
clearword(int y, int x, size_t length)
{
	move(y, x);
	while (length--)
		addch(' ');
}

/*
 * Sets rate according to current length of all visible words.
 * If words with too few chars are on screen, throw in another word.
 */
static void
cwords(void)
{
	int i, length;

	for (length = 0, i = 0; i < 22; i++)
		if (wordpos[i] > 0)
			length += strlen(wordstring[i]) + 1;
	if (length < now.score / 4 + 1) {
		addnewword(NULL);
		rate = now.score / 175 + 3;
	}
}

int
main(int argc, char **argv)
{
	int i, wheretogo;
	char serv[MAXHOSTNAMELEN];

	if (setlocale(LC_CTYPE, "") == NULL)
		errx(1, "Cannot set locale");

	opt.cheat = DEFAULT_CHEAT;
	opt.training = DEFAULT_TRAINING;
	opt.traspeed = DEFAULT_TRAININGSPEED;
	opt.usecolors = DEFAULT_COLORS;
	opt.server = DEFAULT_SERVER;
	opt.client = DEFAULT_CLIENT;
	opt.port = DEFAULT_PORT;

	memset(&best, 0, sizeof(best));
	memset(&now, 0, sizeof(now));
	memset(&word, 0, sizeof(word));

	readconfig();
	printf("Typespeed %s, Copyright (C) Jani Ollikainen & Jaakko Manelius\n", TVERSION);
	puts("                  Copyright (C) Tobias Stoeckmann");
	puts("Typespeed comes with ABSOLUTELY NO WARRANTY; for details read COPYING.\n");

	while ((i = getopt_long(argc, argv, "to:h", options, NULL)) != -1) {
		switch (i) {
		case 0:
			/* flag already set by getopt */
			break;
		case 'm':
			makescorefiles();
			return 0;
			break;
		case 'o':
			opt.client = 1;
			if (!optarg)
				/* getopt should handle this */
				abort();
			(void)strncpy(serv, optarg, sizeof(serv) - 1);
			serv[sizeof(serv) - 1] = '\0';
			break;
		case 'p':
			if (optarg)
				opt.port = atoi(optarg);
			if (opt.port <= 1024) {
				printf("DON'T USE PRIVILEGED PORT\n");
				return 1;
			}
			break;
		case 't':
			opt.training = 1;
			if (optarg)
				opt.traspeed = atoi(optarg);
			if (opt.traspeed < 2 || opt.traspeed > 10)
				opt.traspeed = 2;
			break;
		case ':':
		case '?':
			/* getopt already printed error */
		case 'h':
		default:
			usage(argv[0]);
			break;
		}
	}

	if ((opt.server || opt.client) && opt.training)
		opt.training = 0;

	if (opt.server || opt.client)
		while ((typesocket = initnetwork(serv, 0)) < 0);

	sleep(1);

	initcursestuff();

	if (typesocket > 0)
		play();

	drawmenu();
	/*
	 * THIS IS OLD NEWS - LEFT HERE TO PUT SOME FUTURE NEWS IN
	 * mvaddstr(15,20, "NOTE THIS VERSION HAS NEW WAY TO CALCULATE SPEED!");
	 * mvaddstr(16,26, "SCORES ARE NOT THE SAME AS THEY WERE");
	 */

	wheretogo = 0;
	do {
		nodelay(stdscr, 0);
		cbreak();
	
		while ((wheretogo = getch()) == ERR)
			;

		switch (wheretogo) {
		case '1':
			play();
			break;
		case '2':
			if (opt.training) {
				opt.training = 0;
				drawscreen();
			}
			multipmenu();
			if (typesocket > 0)
				play();
			break;
		case '3':
			tellstory();
			break;
		case '4':
			showhighscores();
			break;
		case '5':
			optionmenu();
			break;
		default:
			break;
		}
		if (wheretogo > '0' && wheretogo < '6')
			drawmenu();
	} while (wheretogo != '6');

	endcursestuff();

	printf("\e[H\e[J");
	printf("Typespeed %s\n\n", TVERSION);
	printf("Best score was: \n\n");
	printf("Rank:\t\t%s\n", rankki[best.level]);
	printf("Score:\t\t%d\n", best.score);
	printf("Total CPS:\t%2.3f\n", best.totalspeed);
	printf("Correct CPS:\t%2.3f\n", best.speed);
	printf("Typo ratio:\t%2.1f%%\n", best.ratio * 100);
	printf("Typorank:\t");
	if (best.score)
		puts(typorank[typorankkaus(best.ratio * 100)]);
	else
		puts("None");
	puts("\n");
	return 0;
}

/*
 * Move words across the screen.
 * Removes them if they reach the right side and increases misses.
 */
static void
movewords(void)
{
	int i;
	int length;

	for (i = 0; i < 22; i++) {
		length = (int)strlen(wordstring[i]);
		xcolor_set(0);
		mvaddch(i, wordpos[i], ' ');

		if (wordpos[i] > (65 - length))
			xcolor_set(3);
		else if (wordpos[i] > (50 - length))
			xcolor_set(7);
		else
			xcolor_set(1);

		if (wordpos[i] > (79 - length)) {
			clearword(i, wordpos[i], length);
			misses++;
			wordpos[i] = 0;
		} else if (wordpos[i] > 0)
			mvprintw(i, ++wordpos[i], "%s", wordstring[i]);
	}
}

static inline void
parseinput(int *escend, char *input, clock_t *starttime, int *inputpos, int *nappaykset, int *wordswritten)
{
	clock_t pausetime;
	int i, nappi;

	if ((nappi = getch()) == ERR)
		return;

	switch (tolower(nappi)) {
	case 8: /* BACKSPACE */
	case 127: /* DEL */
		if (*inputpos > 0) {
			mvaddch(23, 1 + *inputpos, ' ');
			input[--(*inputpos)] = '\0';
		}
		break;
	case 27: /* ESC (sequence) */
		*escend = 1;
		if (opt.client || opt.server)
			break;
		for (i = 0; i < 3; i++)
			/* CURSOR UP is legal */
			if (getch() == 65)
				*escend = 0;
		if (*escend)
			break;
		pausetime = timenow();
		mvaddstr(23, 2, "       PAUSED      ");
		move(23, 2);
		while (getch() == ERR)
			;
		*starttime += timenow() - pausetime;
		mvaddstr(23, 9, "      ");
		flushinp();
		/* FALLTHROUGH */
	case 21: /* ^U */
		if (*inputpos > 0) {
			clearword(23, 2, *inputpos);
			*inputpos = 0;
			input[0] = '\0';
		}
		break;
	case 32: /* SPACE */
	case 10: /* ENTER */
		(*nappaykset)++;
		for (i = 0; i < 22; i++) {
			if (strcmp(input, wordstring[i]))
				continue;
			now.score += *inputpos;
			(*wordswritten)++;
			clearword(i, wordpos[i], strlen(wordstring[i]));
			wordpos[i] = 0;
			/* send ~ 50% words in multiplayer */
			if ((opt.server || opt.client) && !r(2))
				netsendword(typesocket, wordstring[i]);
			(void)strncpy(wordstring[i], " ", sizeof(wordstring[i]) - 1);
		}
		clearword(23, 2, 19);
		*inputpos = 0;
		input[0] = '\0';
		break;
	case 11: /* ^K */
		if (*inputpos < (int)strlen(input)) {
			input[*inputpos] = '\0';
			clearword(23, 2 + *inputpos, 19 - *inputpos);
		}
		break;
	case 6: /* ^F */
		if (*inputpos < 19 && *inputpos < (int)strlen(input))
			(*inputpos)++;
		break;
	case 5: /* ^E */
		if (*inputpos != (int)strlen(input))
			*inputpos = strlen(input);
		break;
	case 2: /* ^B */
		if (*inputpos > 0)
			(*inputpos)--;
		break;
	case 1: /* ^A */
		if (*inputpos > 0)
			*inputpos = 0;
		break;
	default:
		(*nappaykset)++;
		if (*inputpos > 18 || iscntrl(nappi))
			break;
		mvaddch(23, 2 + *inputpos, nappi);
		if (*inputpos == (int)strlen(input))
			input[*inputpos + 1] = '\0';
		input[(*inputpos)++] = nappi;
		if (!opt.cheat)
			break;
		for (i = 0; i < 22; i++) {
			if (strcmp(input, wordstring[i]))
				continue;
			now.score += *inputpos;
			clearword(i, wordpos[i], strlen(wordstring[i]));
			wordpos[i] = 0;
			clearword(23, 2, 19);
			*inputpos = 0;
			input[0]= '\0';
			(void)strncpy(wordstring[i], " ", sizeof(wordstring[i]) - 1);
		}
		break;
	}
	move(23, 2 + *inputpos);
}

static int
play(void)
{
	char input[20];
	clock_t oldtimes, pausetime, starttime;
	float min10;
	int escend, i, inputpos, nappaykset;
	int tra, wordswritten;

	srandom((unsigned)time(NULL));

	if ((opt.client || opt.server) && opt.training)
		opt.training = 0;

	while (choosewordfile(0))
		;

	drawscreen();
	nodelay(stdscr, 1);
	cbreak();

	xcolor_set(2);

	if (opt.server || opt.client) {
		mvaddstr(12, 20, "Waiting for the other party to join in...");
		refresh();
		netwaitgame(typesocket);
		flushinp();
		drawscreen();
	}

	misses = 0;
	min10 = 0;
	escend = inputpos = nappaykset = tra = wordswritten = 0;
	now.score = 0;
	rate = 1;
	now.ratio = 0;
	pausetime = 0;

	memset(wordpos, 0, sizeof(wordpos));

	halfdelay(1);

	oldtimes = starttime = timenow();
	input[0] = '\0';

	for (i = 0; i < 22; i++)
		(void)strncpy(wordstring[i], " ", sizeof(wordstring[i]) - 1);

	while (misses < 10 && escend < 1) {
		while (timenow() < (oldtimes + (100.0 / rate)))
			parseinput(&escend, input, &starttime, &inputpos, &nappaykset, &wordswritten);
		oldtimes = timenow();

		if (opt.training) {
			if (tra == opt.traspeed) {
				movewords();
				cwords();
				tra = 0;
			}
			tra++;
		}
		else {
			movewords();
			cwords();
		}

		now.speed = (now.score + wordswritten) * 100.0 / (timenow() - starttime);
		now.totalspeed = nappaykset * 100.0 / (timenow() - starttime);

		drawstatus(inputpos);
		xcolor_set(2);

		/* last, we check if the opponent has new words for us... */
		if (opt.server || opt.client)
			escend = netrecvword(typesocket);

		refresh();
	}

	if ((opt.server || opt.client) && escend != 2)
		netsendword(typesocket, " ENDOFTHEGAME\n");

	endcursestuff();
	initcursestuff();
	drawscreen();
	now.level = level(now.score);
	if (now.level >= best.level && now.score >= best.score) {
		best.level = now.level;
		best.score = now.score;
		best.speed = now.speed;
		best.totalspeed = now.totalspeed;
		if (now.totalspeed && now.speed)
			best.ratio = (now.totalspeed - now.speed) / now.totalspeed;
	}

	if (opt.server || opt.client)
		netswapscore(typesocket, &now, &other);

	xcolor_set(5);
	mvaddstr(12, 30, "GAME OVER!");
	refresh();

	sleep(3);

	/* total speed */
	min10 = now.speed * 600;

	mvaddstr(12, 30, "          ");
	mvprintw(3, 20, "Typespeed %s ", TVERSION);

	mvaddstr(5, 20, "Your score was: ");
	if (opt.client||opt.server)
		mvaddstr(5, 60, "Opponent's: ");
	mvaddstr(7, 20, "Rank:\t\t");
	/* level(score); */
	mvaddstr(7, 40, rankki[level(now.score)]);
	xcolor_set(5);
	mvprintw(8, 20, "Score:\t\t%d", now.score);
	mvprintw(9, 20, "10MRS:\t\t%1.0f", min10);
	mvprintw(10, 20, "Total CPS:\t\t%2.3f  ", now.totalspeed);
	mvprintw(11, 20, "Correct CPS:\t%2.3f  ", now.speed);
	if (now.totalspeed && now.speed)
		now.ratio = (now.totalspeed - now.speed) / now.totalspeed * 100;
	mvprintw(12, 20, "Typo ratio:\t\t%2.1f%%\t", now.ratio);
	mvaddstr(13, 20, "Typorank:\t\t");

	if (now.score)
		mvaddstr(13, 40, typorank[typorankkaus(now.ratio)]);
	else
		mvaddstr(13, 40, "None");

	if (opt.server || opt.client) {
		mvaddstr(7, 60, rankki[level(other.score)]);
		xcolor_set(5);
		mvprintw(8, 60, "%d", other.score);
		mvprintw(9, 60, "%1.0f", other.speed * 600);
		mvprintw(10, 60, "%2.3f  ", other.totalspeed);
		mvprintw(11, 60, "%2.3f  ", other.speed);
		if (other.totalspeed && other.speed)
			other.ratio = (other.totalspeed - other.speed) / other.totalspeed * 100;
		mvprintw(12, 60, "%2.1f%%", other.ratio);
		if (other.score)
			mvaddstr(13, 60, typorank[typorankkaus(other.ratio)]);
		else
			mvaddstr(13, 60, "None");
	}

	xcolor_set(2);
	if (opt.training)
		mvaddstr(17, 15, "TRAINING MODE - YOU CANNOT GET TO TOP10");
	else if (opt.client || opt.server)
		mvaddstr(17, 20, "NETWORK MODE - YOU CANNOT GET TO TOP10");
	nodelay(stdscr, 0);
	cbreak();
	flushinp();

	pressanykey(15, 20);
	if (!opt.training && !opt.client && !opt.server)
		addtop10(usedwordfile, &now);
	loadhighscores(usedwordfile);

	pressanykey(16, 3);
	free(usedwordfile);
	freewords();
	return 0;
}

/* Prints usage information on stderr and exits with return value 1. */
static void
usage(char *prog)
{
	fputs("\nCommand line options:\n", stderr);
	fputs("  --cheat         = multimedia cheat(?)\n", stderr);
	fputs("  --client=addr   = IP address of server\n", stderr);
	fputs("  --makescores    = Reset/Make new scorefiles\n", stderr);
	fputs("  --nocolors      = don't use colors\n", stderr);
	fputs("  --port=port     = Port number in the network play\n", stderr);
	fputs("  --server        = start typespeed in server mode\n", stderr);
	fputs("  --training[=nm] = training mode\n", stderr);
	fputs("                    (where nm is speed number)\n", stderr);
	fputs("  --help          = this help!\n\n", stderr);
	fprintf(stderr, "Usage: %s options\n", prog);
	exit(1);
}

